@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2009-2022 Free Software Foundation, Inc.
+   Copyright (C) 2009-2025 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
@@ -168,9 +168,15 @@ start_program_filter (int *p, struct _mu_prog_stream *fs, int flags)
   int rc = 0;
   
   if (REDIRECT_STDIN_P (flags))
-    pipe (leftp);
+    {
+      if (pipe (leftp))
+	return errno;
+    }
   if (REDIRECT_STDOUT_P (flags))
-    pipe (rightp);
+    {
+      if (pipe (rightp))
+	return errno;
+    }
   
   switch (fs->pid = fork ())
     {
