@@ -1,3 +1,20 @@
+/* GNU Mailutils -- a suite of utilities for electronic mail
+   Copyright (C) 2003-2025 Free Software Foundation, Inc.
+
+   This library is free software; you can redistribute it and/or
+   modify it under the terms of the GNU Lesser General Public
+   License as published by the Free Software Foundation; either
+   version 3 of the License, or (at your option) any later version.
+
+   This library is distributed in the hope that it will be useful,
+   but WITHOUT ANY WARRANTY; without even the implied warranty of
+   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+   Lesser General Public License for more details.
+
+   You should have received a copy of the GNU Lesser General
+   Public License along with this library.  If not, see
+   <http://www.gnu.org/licenses/>. */
+
 /* TLS I/O streams */
 
 #if HAVE_CONFIG_H
@@ -37,16 +54,43 @@ _tls_io_read (struct _mu_stream *stream, char *buf, size_t bufsize,
 	      size_t *pnread)
 {
   struct _mu_tls_io_stream *sp = (struct _mu_tls_io_stream *) stream;
-  int rc;
+  ssize_t rc;
+  mu_transport_t t[2];
   
   if (sp->up->state != state_open)
     return EINVAL;
-  rc = gnutls_record_recv (sp->up->session, buf, bufsize);
+  do
+    rc = gnutls_record_recv (sp->up->session, buf, bufsize);
+  while (rc == GNUTLS_E_AGAIN || rc == GNUTLS_E_INTERRUPTED);
   if (rc >= 0)
     {
       *pnread = rc;
       return 0;
     }
+
+  switch (rc)
+    {
+    case GNUTLS_E_PUSH_ERROR:
+      if (mu_stream_ioctl (sp->up->transport[1],
+			   MU_IOCTL_TRANSPORT, MU_IOCTL_OP_GET,
+			   &t) == 0 &&
+	  mu_stream_err (t[1]))
+	rc = mu_stream_last_error (t[1]);
+      else
+	rc = MU_ERR_WRITE;
+      return rc;
+      
+    case GNUTLS_E_PULL_ERROR:
+      if (mu_stream_ioctl (sp->up->transport[0],
+			   MU_IOCTL_TRANSPORT, MU_IOCTL_OP_GET,
+			   &t) == 0 &&
+	  mu_stream_err (t[0]))
+	rc = mu_stream_last_error (t[0]);
+      else
+	rc = MU_ERR_READ;
+      return rc;
+    }
+
   sp->up->tls_err = rc;
   return EIO;
 }
@@ -56,7 +100,7 @@ _tls_io_write (struct _mu_stream *stream, const char *buf, size_t bufsize,
 	    size_t *pnwrite)
 {
   struct _mu_tls_io_stream *sp = (struct _mu_tls_io_stream *) stream;
-  int rc;
+  ssize_t rc;
   
   if (sp->up->state != state_open)
     return EINVAL;
@@ -179,6 +223,18 @@ _tls_io_ioctl (struct _mu_stream *stream, int code, int opcode, void *arg)
   return 0;
 }
 
+static void
+_tls_io_event (struct _mu_stream *str, int code,
+	       unsigned long lval, void *pval)
+{
+  struct _mu_tls_io_stream *sp = (struct _mu_tls_io_stream *)str;
+  if (code == _MU_STR_EVENT_CLRFLAG && lval == _MU_STR_ERR)
+    {
+      if (sp->transport)
+	mu_stream_clearerr (sp->transport);
+    }
+}
+
 int
 mu_tls_io_stream_create (mu_stream_t *pstream,
 			 mu_stream_t transport, int flags,
@@ -207,7 +263,9 @@ mu_tls_io_stream_create (mu_stream_t *pstream,
   sp->stream.close = _tls_io_close;
   sp->stream.done = _tls_io_done; 
   sp->stream.ctl = _tls_io_ioctl;
-
+  sp->stream.event_cb = _tls_io_event;
+  sp->stream.event_mask = _MU_STR_EVMASK (_MU_STR_EVENT_CLRFLAG);
+  
   mu_stream_ref (transport);
   sp->transport = transport;
   sp->up = master;
