@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 2019-2021 Free Software Foundation, Inc.
+   Copyright (C) 2019-2025 Free Software Foundation, Inc.
 
    This library is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published by
@@ -72,37 +72,17 @@ mboxrd_mailbox_init_stream (struct mu_mboxrd_mailbox *dmp)
   int rc;
   mu_mailbox_t mailbox = dmp->mailbox;
 
-  /*
-   * Initialize stream flags.  If append mode is requested, convert it to
-   * read-write, so that mboxrd_flush_unlocked be able to update the
-   * X-IMAPbase header in the first message, if necessary.
-   */
-  dmp->stream_flags = mailbox->flags;
-  if (dmp->stream_flags & MU_STREAM_APPEND)
-    dmp->stream_flags = (dmp->stream_flags & ~MU_STREAM_APPEND) | MU_STREAM_RDWR;
-  else if (dmp->stream_flags & MU_STREAM_WRITE)
-    dmp->stream_flags |= MU_STREAM_READ;
-  
-  rc = mu_mapfile_stream_create (&mailbox->stream, dmp->name, dmp->stream_flags);
+  rc = mu_mailbox_stream_create (&mailbox->stream, dmp->name, mailbox->flags);
   if (rc)
     {
       mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
 		("%s:%s (%s): %s",
-		 __func__, "mu_mapfile_stream_create", dmp->name,
+		 __func__, "mu_mailbox_file_open", dmp->name,
 		 mu_strerror (rc)));
-
-      /* Fallback to regular file stream */
-      rc = mu_file_stream_create (&mailbox->stream, dmp->name, dmp->stream_flags);
-      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
-		("%s:%s (%s): %s",
-		 __func__, "mu_file_stream_create", dmp->name,
-		 mu_strerror (rc)));
-
-      if (rc)
-	return rc;
+      return rc;
     }
-
   mu_stream_set_buffer (mailbox->stream, mu_buffer_full, 0);
+  mu_stream_get_flags (mailbox->stream, &dmp->stream_flags);
   return 0;
 }
 
@@ -126,11 +106,11 @@ mboxrd_open (mu_mailbox_t mailbox, int flags)
       && mailbox->locker == NULL
       && (flags & (MU_STREAM_WRITE | MU_STREAM_APPEND | MU_STREAM_CREAT)))
     {
-      rc = mu_locker_create (&mailbox->locker, dmp->name, 0);
+      rc = mu_locker_create_ext (&mailbox->locker, dmp->name, NULL);
       if (rc)
 	mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
 		  ("%s:%s (%s): %s",
-		   __func__, "mu_locker_create", dmp->name,
+		   __func__, "mu_locker_create_ext", dmp->name,
 		   mu_strerror (rc)));
     }
 
@@ -345,7 +325,7 @@ mboxrd_message_alloc_uid (struct mu_mboxrd_message *dmsg)
  * parameters, correspondingly.
  *
  * The header is stored in the first message.  To avoid rewriting entire
- * mailbox when one of the parameters chages, the values of <V> and <N>
+ * mailbox when one of the parameters changes, the values of <V> and <N>
  * are left-padded with spaces to the maximum width of their data types.
  *
  * Offset of the header in the mailbox and its length (without the
@@ -475,7 +455,7 @@ scan_message_finalize (struct mu_mboxrd_mailbox *dmp,
 		 mu_strerror (rc)));
       return -1;
     }
-  dmsg->message_end -= n + 1;
+  dmsg->message_end -= n;
   if (dmsg->uid == 0)
     *force_init_uids = 1;
   if (*force_init_uids)
@@ -583,17 +563,14 @@ scan_message_begin (struct mu_mboxrd_mailbox *dmp, mu_stream_t stream,
  *     different mailbox formats (mboxo, mboxrd and mboxcl (mboxcl2) are
  *     all handled properly.
  *
- *  3. The From_ lines and the blank lines bracket messages, fore and aft.
- *     They do not comprise a message divider.  This means, in particular,
- *     that both bracketing lines should be included in the message octet
- *     and line counts.  However, counting the From_ line goes against
- *     the mailutils approach of logically dividing envelope and the rest
- *     of the message and would create useless differences compared to
- *     another mailbox formats.  For this reason, the From_ line is not
- *     reflected in returns from the mu_message_size and mu_message_lines
- *     functions.  The terminating blank line, on the contrary, is assumed
- *     to be part of the message body and is counted in body and message
- *     size and line count computations.
+ *  3. The From_ line and the last blank line bracket a message, fore and aft.
+ *     For the purpose of internal addressing, both lines are regarded as
+ *     parts of the message, together with message headers and body.  The
+ *     message_start member keeps the offset of the 'F' character in the
+ *     From_ line, and the message_end member keeps the offset of the newline
+ *     character in the terminating empty line.  Neither line is counted when
+ *     computing line and octet counts of the message itself, its headers or
+ *     its body.
  *
  *  4. A mailbox that contains zero messages contains no lines.
  *
@@ -760,7 +737,7 @@ mboxrd_rescan_unlocked (mu_mailbox_t mailbox, mu_off_t offset)
 	case mboxrd_scan_empty_line:
 	  if ((ti = parse_from_line (buf, &zn)) != 0)
 	    {
-	      if (scan_message_finalize (dmp, dmsg, stream, n, &force_init_uids))
+	      if (scan_message_finalize (dmp, dmsg, stream, n + 1, &force_init_uids))
 		goto err;
 	      if ((dmsg = scan_message_begin (dmp, stream, buf, n, ti, zn)) == NULL)
 		goto err;
@@ -777,7 +754,9 @@ mboxrd_rescan_unlocked (mu_mailbox_t mailbox, mu_off_t offset)
 
   if (dmsg)
     {
-      if (scan_message_finalize (dmp, dmsg, stream, 0, &force_init_uids))
+      if (scan_message_finalize (dmp, dmsg, stream,
+				 state == mboxrd_scan_empty_line ? 1 : 0,
+				 &force_init_uids))
 	goto err;
     }
   
@@ -790,7 +769,8 @@ mboxrd_rescan_unlocked (mu_mailbox_t mailbox, mu_off_t offset)
 		 mu_strerror (rc)));
     }
   mu_stream_unref (stream);
-
+  free (buf);
+  
   if (force_init_uids)
     {
       dmp->uidvalidity = (unsigned long) time (NULL);
@@ -936,7 +916,8 @@ mboxrd_message_unseen (mu_mailbox_t mailbox, size_t *pmsgno)
 	}
     }
 
-  return MU_ERR_NOENT;
+  *pmsgno = 0;
+  return 0;
 }
 
 /* Initialize the mailbox UID subsystem. See the Notes above. */
@@ -1064,7 +1045,8 @@ mboxrd_quick_get_message (mu_mailbox_t mailbox, mu_message_qid_t qid,
 }
 
 static int
-mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
+mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg,
+			mu_envelope_t env, mu_attribute_t atr)
 {
   int rc;
   mu_off_t size;
@@ -1072,45 +1054,59 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
   static char *exclude_headers[] = {
     MU_HEADER_X_IMAPBASE,
     MU_HEADER_X_UID,
+    MU_HEADER_STATUS,
     NULL
   };
   struct mu_mboxrd_mailbox *dmp = mailbox->data;
-
+  
   if (dmp->mesg_count)
     {
       char nl[2];
       static char pad[] = { '\n', '\n' };
       int n;
+      size_t nread;
       
       size = dmp->mesg[dmp->mesg_count-1]->message_end - 1;
       rc = mu_stream_seek (mailbox->stream, size, MU_SEEK_SET, NULL);
       if (rc)
 	return rc;
-      rc = mu_stream_read (mailbox->stream, nl, 2, NULL);
+      rc = mu_stream_read (mailbox->stream, nl, 2, &nread);
       if (rc)
 	return rc;
 
-      if (nl[1] != '\n')
-	n = 2;
-      else if (nl[0] != '\n')
-	n = 1;
+      if (nread == 2)
+	{
+	  if (nl[1] != '\n')
+	    n = 2;
+	  else if (nl[0] != '\n')
+	    n = 1;
+	  else
+	    n = 0;
+	}
+      else if (nread == 1)
+	{
+	  if (nl[0] != '\n')
+	    n = 2;
+	  else
+	    n = 1;
+	}
       else
-	n = 0;
-
+	n = 2;
+      
       if (n)
 	{
 	  mu_stream_write (mailbox->stream, pad, n, NULL);
+	  dmp->mesg[dmp->mesg_count-1]->message_end += n - 1;
 	}
-      size += n + 2;
+      size += n + nread;
     }
   else
     {
       size = 0;
       rc = mu_stream_seek (mailbox->stream, size, MU_SEEK_SET, NULL);
+      if (rc)
+	return rc;
     }
-  
-  if (rc)
-    return rc;
 
   rc = mu_message_get_streamref (msg, &istr);
   if (rc)
@@ -1118,14 +1114,17 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
 
   do
     {
-      mu_envelope_t env;
       char *date = NULL;
       char *sender = NULL;
+      char statbuf[MU_STATUS_BUF_SIZE];
 
-      rc = mu_message_get_envelope (msg, &env);
-      if (rc)
-	break;
-
+      if (!env)
+	{
+	  rc = mu_message_get_envelope (msg, &env);
+	  if (rc)
+	    break;
+	}
+      
       rc = mu_envelope_aget_sender (env, &sender);
       if (rc == 0)
 	{
@@ -1139,6 +1138,14 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
 		  if (rc == 0)
 		    {
 		      rc = mu_envelope_aget_date (env, &date);
+		      if (rc == MU_ERR_NOENT)
+			{
+			  date = strdup ("Thu Jan  1 00:00:00 1970");
+			  if (date)
+			    rc = 0;
+			  else
+			    rc = ENOMEM;
+			}
 		    }
 
 		  mu_envelope_destroy (&env, msg);
@@ -1159,11 +1166,25 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
 
       if (rc)
 	break;
+
+      if (atr)
+	{
+	  rc = mu_attribute_to_string (atr, statbuf, MU_STATUS_BUF_SIZE, NULL);
+	  if (rc)
+	    break;
+	}
+      else
+	statbuf[0] = 0;
       
       rc = mu_stream_header_copy (mailbox->stream, istr, exclude_headers);
       if (rc)
 	break;
 
+      /* Write status header */
+      if (statbuf[0])
+	mu_stream_printf (mailbox->stream,
+			  "%s: %s\n", MU_HEADER_STATUS, statbuf);
+      
       /* Write UID-related data */
       if (dmp->uidvalidity_scanned)
 	{
@@ -1183,7 +1204,7 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
       if (rc)
 	break;
 
-      rc = mu_filter_create (&flt, istr, "FROMRB",
+      rc = mu_filter_create (&flt, istr, "FROMRD",
 			     MU_FILTER_ENCODE, MU_STREAM_READ);
       mu_stream_destroy (&istr);
       rc = mu_stream_copy_nl (mailbox->stream, flt, 0, NULL);
@@ -1193,11 +1214,12 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
 
   if (rc)
     {
+      int rc1;
       mu_stream_destroy (&istr);
-      rc = mu_stream_truncate (mailbox->stream, size);
-      if (rc)
+      rc1 = mu_stream_truncate (mailbox->stream, size);
+      if (rc1)
 	mu_error (_("cannot truncate stream after failed append: %s"),
-		  mu_stream_strerror (mailbox->stream, rc));
+		  mu_stream_strerror (mailbox->stream, rc1));
       return rc;
     }
 
@@ -1219,7 +1241,8 @@ mailbox_append_message (mu_mailbox_t mailbox, mu_message_t msg)
 }
 
 static int
-mboxrd_append_message (mu_mailbox_t mailbox, mu_message_t msg)
+mboxrd_append_message (mu_mailbox_t mailbox, mu_message_t msg,
+		       mu_envelope_t env, mu_attribute_t atr)
 {
   struct mu_mboxrd_mailbox *dmp = mailbox->data;
   int rc;
@@ -1238,7 +1261,7 @@ mboxrd_append_message (mu_mailbox_t mailbox, mu_message_t msg)
     }
   else
     {
-      rc = mailbox_append_message (mailbox, msg);
+      rc = mailbox_append_message (mailbox, msg, env, atr);
 
       if (mailbox->locker)
 	mu_locker_unlock (mailbox->locker);
@@ -1280,18 +1303,63 @@ mboxrd_get_size (mu_mailbox_t mailbox, mu_off_t *psize)
   return 0;
 }
 
+static int
+mboxrd_stat (mu_mailbox_t mailbox, struct stat *st)
+{
+  int rc;
+  mu_transport_t trans[2];
+  
+  rc = mu_stream_ioctl (mailbox->stream, MU_IOCTL_TRANSPORT,
+			MU_IOCTL_OP_GET, trans);
+  if (rc == 0)
+    {
+      if (fstat ((int) (intptr_t) trans[0], st))
+	rc = errno;
+    }
+  return rc;
+}
+
+static int
+mboxrd_set_priv (struct mu_mboxrd_mailbox *dmp, struct stat *st)
+{
+  int rc;
+  mu_transport_t trans[2];
+  
+  rc = mu_stream_ioctl (dmp->mailbox->stream, MU_IOCTL_TRANSPORT,
+			MU_IOCTL_OP_GET, trans);
+  if (rc == 0)
+    {
+      int fd = (intptr_t) trans[0];
+      if (fchmod (fd, st->st_mode))
+	{
+	  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+		    ("%s:%s: chmod failed: %s",
+		     __func__, dmp->name, strerror (errno)));
+	  rc = errno;
+	}
+      else if (fchown (fd, st->st_uid, st->st_gid))
+	{
+	  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+		    ("%s:%s: chown failed: %s",
+		     __func__, dmp->name, strerror (errno)));
+	  rc = errno;
+	}
+    }
+  return rc;
+}
+
 static int
 mboxrd_get_atime (mu_mailbox_t mailbox, time_t *return_time)
 {
   struct mu_mboxrd_mailbox *dmp = mailbox->data;
   struct stat st;
-
+  int rc;
+  
   if (dmp == NULL)
     return EINVAL;
-  if (stat (dmp->name, &st))
-    return errno;
-  *return_time = st.st_atime;
-  return 0;
+  if ((rc = mboxrd_stat (mailbox, &st)) == 0)
+    *return_time = st.st_atime;
+  return rc;
 }
 
 struct mu_mboxrd_flush_tracker
@@ -1338,6 +1406,9 @@ mboxrd_tracker_sync (struct mu_mboxrd_flush_tracker *trk)
 	mu_mboxrd_message_free (dmp->mesg[i]);      
       dmp->size = 0;
       dmp->uidvalidity_scanned = 0;
+      dmp->uidvalidity_changed = 0;
+      dmp->x_imapbase_off = 0;
+      dmp->x_imapbase_len = 0;
     }
   else
     {
@@ -1496,6 +1567,54 @@ mboxrd_flush_temp (struct mu_mboxrd_flush_tracker *trk,
   return mu_stream_flush (tempstr);
 }
 
+/*
+ * Copy the temporary mailbox stream TEMPSTR to the mailbox referred to by
+ * the tracker TRK.
+ */
+static inline int
+mboxrd_copyback (struct mu_mboxrd_flush_tracker *trk, mu_stream_t tempstr)
+{
+  int rc;
+  mu_stream_t mbx_stream = trk->dmp->mailbox->stream;
+  mu_off_t size;
+  
+  rc = mu_stream_seek (tempstr, 0, MU_SEEK_SET, NULL);
+  if (rc)
+    {
+      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+		("%s: can't rewind temporary file: %s",
+		 __func__, mu_strerror (rc)));
+      return rc;
+    }
+
+  rc = mu_stream_seek (mbx_stream, 0, MU_SEEK_SET, NULL);
+  if (rc)
+    {
+      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+		("%s: can't rewind mailbox %s: %s",
+		 __func__, trk->dmp->name, mu_strerror (rc)));
+      return rc;
+    }
+
+  rc = mu_stream_copy (mbx_stream, tempstr, 0, &size);
+  if (rc)
+    {
+      mu_error (_("copying back to mailbox %s failed: %s"),
+		trk->dmp->name, mu_strerror (rc));
+      return rc;
+    }
+  rc = mu_stream_truncate (mbx_stream, size);
+  if (rc)
+    {
+      mu_error (_("cannot truncate mailbox stream: %s"),
+		mu_stream_strerror (mbx_stream, rc));
+      return rc;
+    }
+  
+  mboxrd_tracker_sync (trk);
+  return 0;
+}
+
 /* Flush the mailbox described by the tracker TRK to the stream TEMPSTR.
    EXPUNGE is 1 if the MU_ATTRIBUTE_DELETED attribute is to be honored.
    Assumes that simultaneous access to the mailbox has been blocked.
@@ -1616,13 +1735,26 @@ mboxrd_flush_unlocked (struct mu_mboxrd_flush_tracker *trk, int mode)
 	    return ENOMEM;
 	}
       rc = mu_tempfile (&hints, MU_TEMPFILE_TMPDIR, &tempfd, &tempname);
-      if (rc)
+      if (rc == 0)
 	{
-	  free (hints.tmpdir);
-	  return rc;
+	  rc = mu_fd_stream_create (&tempstr, tempname, tempfd,
+				    MU_STREAM_RDWR|MU_STREAM_SEEK);
 	}
-      rc = mu_fd_stream_create (&tempstr, tempname, tempfd,
-				MU_STREAM_RDWR|MU_STREAM_SEEK);
+      else if (rc == EACCES)
+	{
+	  /*
+	   * Mail spool directory is not writable for the user. Fall
+	   * back to using temporary stream located elsewhere. When
+	   * ready, it will be copied back to the mailbox.
+	   *
+	   * Reset the tempname to NULL to instruct the code below
+	   * which approach to take.
+	   */
+	  tempname = NULL;
+
+	  rc = mu_temp_file_stream_create (&tempstr, NULL, 0);
+	}
+
       if (rc)
 	{
 	  free (hints.tmpdir);
@@ -1632,58 +1764,83 @@ mboxrd_flush_unlocked (struct mu_mboxrd_flush_tracker *trk, int mode)
 	}
       
       rc = mboxrd_flush_temp (trk, dirty, tempstr, mode == FLUSH_EXPUNGE);
-      mu_stream_unref (tempstr);
       if (rc == 0)
 	{
-	  /* Rename mailbox to temporary copy */
-	  char *backup = mu_tempname (hints.tmpdir);
-	  rc = rename (dmp->name, backup);
-	  if (rc)
+	  if (tempname)
 	    {
-	      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
-			("%s:%s: failed to rename to backup file %s: %s",
-			 __func__, dmp->name, tempname,
-			 mu_strerror (rc)));
-	      unlink (backup);
-	    }
-	  else
-	    {
-	      rc = rename (tempname, dmp->name);
-	      if (rc == 0)
+	      /* Mail spool is writable. Rename the temporary copy back
+		 to mailbox */
+	      char *backup;
+	      struct stat st;
+
+	      if ((rc = mboxrd_stat (dmp->mailbox, &st)) != 0)
 		{
-		  /* Success. Synchronize internal data with the counter. */
-		  mboxrd_tracker_sync (trk);
-		  mu_stream_destroy (&dmp->mailbox->stream);
-		  rc = mboxrd_mailbox_init_stream (dmp);
+		  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+			    ("%s:%s: stat failed: %s",
+			     __func__, dmp->name, strerror (errno)));
 		}
 	      else
 		{
-		  int rc1;
-		  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
-			    ("%s: failed to rename temporary file %s %s: %s",
-			     __func__, tempname, dmp->name,
-			     mu_strerror (rc)));
-		  rc1 = rename (backup, dmp->name);
-		  if (rc1)
+		  mu_stream_flush (tempstr);
+		  backup = mu_tempname (hints.tmpdir);
+		  if (rename (dmp->name, backup))
 		    {
-		      mu_error (_("failed to restore %s from backup %s: %s"),
-				dmp->name, backup, mu_strerror (rc1));
-		      mu_error (_("backup left in %s"), backup);
-		      free (backup);
-		      backup = NULL;
+		      rc = errno;
+		      mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+				("%s:%s: failed to rename to backup file %s: %s",
+				 __func__, dmp->name, tempname,
+				 mu_strerror (rc)));
+		      unlink (backup);
 		    }
+		  else
+		    {
+		      rc = rename (tempname, dmp->name);
+		      if (rc == 0)
+			{
+			  /* Success. Synchronize internal data with the
+			     counter. */
+			  mboxrd_tracker_sync (trk);
+			  mu_stream_destroy (&dmp->mailbox->stream);
+			  rc = mboxrd_mailbox_init_stream (dmp);
+			  if (rc == 0)
+			    mboxrd_set_priv (dmp, &st);
+			}
+		      else
+			{
+			  int rc1;
+			  mu_debug (MU_DEBCAT_MAILBOX, MU_DEBUG_ERROR,
+				    ("%s: failed to rename temporary file %s %s: %s",
+				     __func__, tempname, dmp->name,
+				     mu_strerror (rc)));
+			  rc1 = rename (backup, dmp->name);
+			  if (rc1)
+			    {
+			      mu_error (_("failed to restore %s from backup %s: %s"),
+					dmp->name, backup, mu_strerror (rc1));
+			      mu_error (_("backup left in %s"), backup);
+			      free (backup);
+			      backup = NULL;
+			    }
+			}
+		    }
+	  
+		  if (backup)
+		    {
+		      unlink (backup);
+		      free (backup);
+		    }
+		  unlink (tempname);
 		}
 	    }
-	  
-	  if (backup)
+	  else
 	    {
-	      unlink (backup);
-	      free (backup);
+	      /* Mail spool not writable.  Copy the tempstr back to mailbox. */
+	      rc = mboxrd_copyback (trk, tempstr);
 	    }
 	}
-      unlink (tempname);
       free (tempname);
       free (hints.tmpdir);
+      mu_stream_unref (tempstr);
     }
   
   dmp->uidvalidity_changed = 0;  
@@ -1864,6 +2021,7 @@ mboxrd_detect (char const *name)
 		}
 	    }
 	  free (buf);
+	  mu_stream_destroy (&str);
 	}
     }
   return res;
