@@ -1,5 +1,5 @@
 /* GNU Mailutils -- a suite of utilities for electronic mail
-   Copyright (C) 1999-2022 Free Software Foundation, Inc.
+   Copyright (C) 1999-2025 Free Software Foundation, Inc.
 
    GNU Mailutils is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -145,7 +145,7 @@ union mailvar_value
 {
   char *string;
   int number;
-  int bool;
+  int boolean;
 };
 
 struct mailvar_variable
@@ -479,6 +479,7 @@ int util_getlines (void);
 int util_screen_lines (void);
 int util_screen_columns (void);
 int util_get_crt (void);
+void mailvar_init_variables (void);
 struct mailvar_variable *mailvar_find_variable (const char *var, int create);
 int mailvar_get (void *ptr, const char *variable,
 		 enum mailvar_type type, int warn);
@@ -593,6 +594,7 @@ char *readline (char *prompt);
 #define MSG_NODELETED 0x0001
 #define MSG_SILENT    0x0002
 #define MSG_COUNT     0x0004
+#define MSG_ALLOWPART 0x0008
 
 /* Message attributes */
 #define MAIL_ATTRIBUTE_MBOXED   0x0001
